# coding=utf-8
'''
Created on 18.3.2013

Holds, loads and saves measuring settings.
'''
__author__ = "Jarkko Aalto \n Timo Konu \n Samuli Kärkkäinen \n Samuli Rahkonen \n Miika Raunio"
__versio__ = "1.0"

import configparser
import os
from Modules.Element import Element

class MeasuringSettings:
    """MeasuringSettings holds the all project specific measurement unit parameters.
    """
    def __init__(self, settings_filepath=None):
        """Inits MeasuringSettings.
        
        Args:
            settings_filepath: filepath for the settings file to be loaded
        """
        
        
        self.measuring_unit_settings_filename = "measuring_unit_settings.ini"
        self.config = configparser.ConfigParser()
        # This is used to determine if opened measurement 
        # uses its own settings ('MEASUREMENT') or project's 
        # settings ('PROJECT') 
        self.use_settings = ""         
        
        
        #self.element = ""
        self.element = Element()
        #self.isotope = 0
        self.energy = 0
        self.detector_angle = 0
        self.target_angle = 0
        self.time_of_flight_lenght = 0
        self.carbon_foil_thickness = 0
        self.target_density = 0
        
        
        
        self.__set_config_parameters()
        if settings_filepath == None:
            return
        
        self.filepath = os.path.join(settings_filepath, self.measuring_unit_settings_filename) 
        #if not os.path.exists(settings_filepath):
        if not os.path.exists(self.filepath):
            #print("Creating " + settings_filepath)
            print("Creating " + self.filepath)
            #with open(settings_filepath, 'w+') as configfile:  # save
            with open(self.filepath, 'w+') as configfile:  # save
                self.config.write(configfile)
        
        self.filepath = os.path.join(settings_filepath, self.measuring_unit_settings_filename) 
        if not os.path.exists(settings_filepath):
            print("Creating " + settings_filepath)
            with open(settings_filepath, 'w+') as configfile:  # save
                self.config.write(configfile)
            
        #self.load_settings(settings_filepath)
        self.load_settings(self.filepath)
        
        
    def show(self, dialog):
        """Shows the measuring parameters in the given measuring settings dialog.
        
        Args:
            dialog: Measuring Settings QDialog whose fields are updated with the 
            MeasuringSettings parameters.
        """
        try:
            if self.use_settings == "PROJECT":
                dialog.ui.useProjectSettingsValuesCheckBox.setChecked(True)
            elif self.use_settings == "MEASUREMENT":
                dialog.ui.useProjectSettingsValuesCheckBox.setChecked(False)
        except:
            print("Can't find the checkbox in the dialog")
        """
        try:
            dialog.useProjectSettingsValuesCheckBox.setChecked(True)
        except:
            print("Can't find checkbox") #This happens with project settings dialog.
        """
        dialog.elementButton.setText(self.element.name)
        dialog.energyLineEdit.setText(str(self.energy))
        dialog.detectorAngleLineEdit.setText(str(self.detector_angle))
        dialog.targetAngleLineEdit.setText(str(self.target_angle))
        dialog.TOFLengthLineEdit.setText(str(self.time_of_flight_lenght))
        dialog.carbonFoilThicknessLineEdit.setText(str(self.carbon_foil_thickness))
        dialog.targetDensityLineEdit.setText(str(self.target_density))
        
        
    def set_settings(self, dialog, used_settings=None):
        """Takes inputted parameters from the given dialog and sets them to the 
        corresponding object's parameters
        
        Args:
            dialog: Measuring Settings QDialog from which the parameters are taken.
        """
        
        if used_settings == None:
            self.use_settings = ""
        else:
            self.use_settings = used_settings
        
        self.element = Element(dialog.elementButton.text(), 
                               dialog.isotopeComboBox.currentText())
        #
        #if isotope == "":
        #    self.isotope = 0
        #else:
        #    self.isotope = isotope 
        self.energy = float(dialog.energyLineEdit.text())
        self.detector_angle = float(dialog.detectorAngleLineEdit.text())
        self.target_angle = float(dialog.targetAngleLineEdit.text())
        self.time_of_flight_lenght = float(dialog.TOFLengthLineEdit.text())
        self.carbon_foil_thickness = float(dialog.carbonFoilThicknessLineEdit.text())
        self.target_density = float(dialog.targetDensityLineEdit.text())
        
        
    def load_settings(self, filepath):
        """Loads settings' parameters from the given filepath.
        
        Args:
            filepath: Filepath to the settings file.
        """
        if not os.path.exists(filepath):
            print("Filepath " + filepath + " does not exist.")
            return
        self.config.read(filepath)
        
        # print(self.config['DEFAULT']['path'])     # -> "/path/name/"
        
        try:
            self.use_settings = self.config['default']['use_settings']
        
            #self.element = self.config['beam']['element']
            #isotope = self.config['beam']['isotope']
            self.element = Element(self.config['beam']['element'], self.config['beam']['isotope'])
            #if isotope == "":
            #    self.isotope = 0
            #else:
            #    
            #    self.isotope = float(self.config['beam']['isotope'])
            self.energy = float(self.config['beam']['energy'])
            
            self.detector_angle = float(self.config['measuring_unit']['detector_angle'])
            self.target_angle = float(self.config['measuring_unit']['target_angle'])
            self.time_of_flight_lenght = float(self.config['measuring_unit']['time_of_flight_lenght'])
            self.carbon_foil_thickness = float(self.config['measuring_unit']['carbon_foil_thickness'])
            self.target_density = float(self.config['measuring_unit']['target_density'])
        except:  # If there is a problem, use default values
            print("Couldn't load measurement unit settings from file")
            return
    
        
    def __set_config_parameters(self):
        self.config.add_section('default')
        self.config.set('default', 'use_settings', str(self.use_settings))
        self.config.add_section('beam')
        self.config.set('beam', 'element', self.element.name)
        self.config.set('beam', 'isotope', str(self.element.isotope))
        self.config.set('beam', 'energy', str(self.energy))
        self.config.add_section('measuring_unit')
        self.config.set('measuring_unit', 'detector_angle', str(self.detector_angle))
        self.config.set('measuring_unit', 'target_angle', str(self.target_angle))
        self.config.set('measuring_unit', 'time_of_flight_lenght', str(self.time_of_flight_lenght)) 
        self.config.set('measuring_unit', 'carbon_foil_thickness', str(self.carbon_foil_thickness))
        self.config.set('measuring_unit', 'target_density', str(self.target_density))
    
    
        
    def save_settings(self, filepath=None):
        """Saves settings' parameters to the given filepath.
        
        Args:
            filepath: Filepath to the settings file.
        """
        
        # self.config['DEFAULT']['path'] = 'value'
        self.config['default']['use_settings'] = self.use_settings
        
        self.config['beam']['element'] = str(self.element.name)
        self.config['beam']['isotope'] = str(self.element.isotope)
        self.config['beam']['energy'] = str(self.energy)
        
        self.config['measuring_unit']['detector_angle'] = str(self.detector_angle)
        self.config['measuring_unit']['target_angle'] = str(self.target_angle)
        self.config['measuring_unit']['time_of_flight_lenght'] = str(self.time_of_flight_lenght)
        self.config['measuring_unit']['carbon_foil_thickness'] = str(self.carbon_foil_thickness)
        self.config['measuring_unit']['target_density'] = str(self.target_density)
        
        if filepath == None:
            filepath = self.filepath
        
        with open(filepath, 'w+') as configfile:  # save
            self.config.write(configfile)
        
